#ifndef _FFT_EARS_
#define _FFT_EARS_
#include "fft_view.h"
#include "fft_defs.h"

#include <Subscriber.h>
#include <AudioStream.h>

/* Audio stream settings. */
#define SRATE 44100
#define STREAM_BUF_COUNT 2

/* Don't change the bufsize until the other fix
	tables are in place.
*/

#define STREAM_BUF_SIZE 2048
#define SAMPLE_SIZE 2
#define CHANNEL_COUNT 2

/* Audio processing:

Three threads:
	-ADC (Media Kit)
	-PROC (whisper)
	-DAC (Media Kit)

The ADC thread (Listen) converts 16-bit stereo to float mono.
The proc thread does fft/fiddle/inverse-fft
The DAC thread (Talk) converts float mono to 16-bit stereo.

Confounding all of this are the overlap buffers:

	- To smooth the output, we need "overlap" buffers that are the last
	half of buf N plus the first half of N+1 which are then run
	through the proc thread.
	- On the output side, each dac buffer is the last half of the
	previous overlap buffer ("underlap"), plus this buffer ("current"), 
	plus the first half of the next overlap buffer ("overlap").
	- Each buffer has to be windowed. The float -> int16 routine
	performs the windowing directly by factoring the type conversion
	into the window table.
	
The adc thread creates the overlap buffers.
The dac thread decodes them.

Because of the overlap, we need twice as many sample buffers as
stream buffers.
 */
#define SAMPLE_COUNT (STREAM_BUF_SIZE/(SAMPLE_SIZE*CHANNEL_COUNT))
#define SAMPLE_BUF_COUNT (STREAM_BUF_COUNT*2)
#define SAMPLE_BUF_SIZE (SAMPLE_COUNT*sizeof(float))
#define FFT_BUF_SIZE SAMPLE_BUF_SIZE

#define FFT_BIN_COUNT SAMPLE_COUNT

#define CLOBBER_BIN_COUNT 0

#define COMB_STEPS 8
#define WHISPER_BUF_COUNT 8
#define DELAY_COUNT 15

class fft_ears
{
	public:
		fft_ears(fft_view *v);
		~fft_ears(void);

		void Go();
	
		void SetSRate(float srate);
		void SetFiddleMode(fiddle_mode mode);
		void SetFiddleValue(float fiddle);
		void Fiddle(float *buf, int32 count);

		/* Fiddle functions.  The input buffer is an fft;
		 * it's fiddled in place.
		 */
		void Shift(float *buf, int32 count);
		void Comb(float *buf, int32 count);
		void Convolve(float *buf, int32 count);
		void Whisper(float *buf, int32 count);
		void Flipper(float *buf, int32 count);
		void Splatter(float *buf, int32 count);
	
		void SetBarMode(bar_mode which);
		void SetBars(float *buf, int32 start);
		
	private:

		void InitFiddle();
		void InitGraphicStuff();
		void InitAudioStuff();
		void InitThreadsEtAl();

		void ShiftFiddleValue(float val);
		void CombFiddleValue(float val);
		void WhisperFiddleValue(float val);
		void FlipperFiddleValue(float val);
		void ConvolveFiddleValue(float val);
		void SplatterFiddleValue(float val);

		/* Graphic stuff */
		fft_view *bar_view;
		uint8 bar_buf[FFT_BAR_COUNT];
		int32 look_at;
	
		/* Audio stream stuff */
		BSubscriber *ears;
		BSubscriber *mouth;
		BDACStream *dac;
		BADCStream *adc;
		float srate;
		
		/* Sample buffers.  
		 */
		float adc_hold[SAMPLE_COUNT];
		float samples[SAMPLE_BUF_COUNT][SAMPLE_COUNT];
		float aux[SAMPLE_COUNT];
		float convolve_bufs[SAMPLE_COUNT][DELAY_COUNT];
		uint32 convolve_index;
		
		/* The whisper buffers store magnitude. */
		float whispers[WHISPER_BUF_COUNT][SAMPLE_COUNT/2];
		int32 whisper_index;
		/* The amount to <x> the signal*/
		float fiddle_val;
		float real_val;
		fiddle_mode which_fiddle;
		
		/* Sample write (adc) */
		sem_id write_sems[SAMPLE_BUF_COUNT];
		int32 write_atoms[SAMPLE_BUF_COUNT];
		int32 write_index;

		/* Sample read (dac) */
		sem_id read_sems[SAMPLE_BUF_COUNT];
		int32 read_atoms[SAMPLE_BUF_COUNT];
		int32 read_index;

		thread_id proc_thread;
		bool proc_continue;
		static int32 _proc(void *arg);
		void Proc(void);
			
		static bool _listen(void *data, char *buf, size_t count,
			void *header);
		bool Listen(int16 *data, int32 count);

		static bool _talk(void *data, char *buf, size_t count,
			void *header);
		bool Talk(int16 *data, int32 count);
};
inline void fft_ears::SetBarMode(bar_mode which) { look_at = which; }

#endif